
//utility functions

    //generates frequencies based on the 12 tone equal temperament scale
    //based on a factor of how far we are between two notes.
    //NB: if setting noteRound to false you can return fractional frequencies.
    //based on A (440Hz) being note '49'
    //to just get a single note number -> frequency, call freq12TET(yourNote, yourNote, 0);
    function freq12TET(minNote, maxNote, betweenFactor, noteRound = true)
    {
        //https://en.wikipedia.org/wiki/Equal_temperament#Twelve-tone_equal_temperament

        var P_a = 440.0;
        var a = 49;
        var n = minNote + (maxNote - minNote) * betweenFactor;

        if (noteRound)
        {
            n = Math.round(n);
        }

        var P_n = P_a * Math.pow( Math.pow(2, 1/12) , n - a);

        return P_n;
    }

    //provides some gain levelling
    function freqToGain(frequency)
    {
        return Math.pow(1.0 - (frequency / 20000), 12); //4 //8 //20
    }


//canvas helper methods

    function canvasCircle(ctx, x, y, r, fill = false)
    {
        ctx.beginPath();
        ctx.arc(x, y, r, 0, 2 * Math.PI);
        if (fill)
        {
            ctx.fill();
        }
        else
        {
            ctx.stroke();
        }
    }

    function canvasRGBA(r, g, b, a)
    {
        return "rgba(" + r + ", " + g + ", " + b + ", " + a + ")";
    }


//fundamental web audio api classes

    class Envelope {

        constructor(a = 0.1, d = 0.1, s = 0.5, r = 0.1, m = 1.0)
        {
            this.attack = a;
            this.decay = d;
            this.sustain = s;
            this.release = r;

            this.multiplier = m; //this allows envelopes other than 0-1.
                                //the envelope will attack to multiplier, then decay
                                //to sustain x muliplier.
        }

        applyOn(audioParamGain, context, fromZero = false)
        {
            audioParamGain.cancelScheduledValues(0);
            if (fromZero)
            {
                audioParamGain.setValueAtTime(0.01, context.currentTime);
            }
            else
            {
                audioParamGain.setValueAtTime(audioParamGain.value == 0 ? 0.01 : audioParamGain.value, context.currentTime);
            }
            //attack
            audioParamGain.exponentialRampToValueAtTime(this.multiplier, context.currentTime + this.attack);
            //decay - use setTargetAtTime if there is any sustain, otherwise use exp + final drop to actual 0.
            if (this.sustain > 0)
            {
                audioParamGain.setTargetAtTime(this.sustain * this.multiplier, context.currentTime + this.attack, this.decay / 3);
            }
            else
            {
                audioParamGain.exponentialRampToValueAtTime(0.001, context.currentTime + this.attack + this.decay);
                audioParamGain.linearRampToValueAtTime(0, context.currentTime + this.attack + this.decay + 0.001);
            }
        }

        applyOff(audioParamGain, context)
        {
            audioParamGain.cancelScheduledValues(0);
            audioParamGain.setValueAtTime(audioParamGain.value, context.currentTime);
            audioParamGain.linearRampToValueAtTime(0, context.currentTime + this.release);
        }

    }

    //https://stackoverflow.com/questions/34482319/web-audio-api-how-do-i-add-a-working-convolver
    function impulseResponse( duration, decay, reverse ) {
        var sampleRate = audioContext.sampleRate;
        var length = sampleRate * duration;
        var impulse = audioContext.createBuffer(2, length, sampleRate);
        var impulseL = impulse.getChannelData(0);
        var impulseR = impulse.getChannelData(1);

        if (!decay)
            decay = 2.0;
        for (var i = 0; i < length; i++){
          var n = reverse ? length - i : i;
          impulseL[i] = (Math.random() * 2 - 1) * Math.pow(1 - n / length, decay);
          impulseR[i] = (Math.random() * 2 - 1) * Math.pow(1 - n / length, decay);
        }
        return impulse;
    }

//fundamental game stuff

// https://stackoverflow.com/questions/9043805/test-if-two-lines-intersect-javascript-function
// returns true if the line from (a,b)->(c,d) intersects with (p,q)->(r,s)
function intersects(a,b,c,d,p,q,r,s) {
  var det, gamma, lambda;
  det = (c - a) * (s - q) - (r - p) * (d - b);
  if (det === 0) {
    return false;
  } else {
    lambda = ((s - q) * (r - a) + (p - r) * (s - b)) / det;
    gamma = ((b - d) * (r - a) + (c - a) * (s - b)) / det;
    return (0 < lambda && lambda < 1) && (0 < gamma && gamma < 1);
  }
};

    class Vector
    {
    	constructor(dx, dy)
    	{
    		this.dx = dx;
    		this.dy = dy;

    		//we will cache the vector size to reduce sqrt calcs.
    		this.sizeCache = 0;
    		this.sizeCacheReady = false;
    	}

    	set(dx, dy)
    	{
    		this.dx = dx;
    		this.dy = dy;
    		this.sizeCacheReady = false;
    	}

        size()
        {
    		if (!this.sizeCacheReady)
    		{
    			this.sizeCache = Math.sqrt(this.dx * this.dx + this.dy * this.dy);
    			this.sizeCacheReady = true;

    		}
            return this.sizeCache;
        }

    	size2()
    	{
    		return (this.dx * this.dx + this.dy * this.dy);
    	}

    	add(vector)
    	{
    		this.dx = this.dx + vector.dx;
            this.dy = this.dy + vector.dy;
    		this.sizeCacheReady = false;
    	}

    	multiply(number)
    	{
    		this.dx = this.dx * number;
            this.dy = this.dy * number;
    		this.sizeCacheReady = false;
    	}

        divide(number)
        {
            if (number != 0)
            {
                this.dx = this.dx / number;
                this.dy = this.dy / number;
    			this.sizeCacheReady = false;
            }

        }

        //class level version updates the instance being called, to get a new vector use the static method Vector.subtract
        subtract(vector)
        {
            this.dx = this.dx - vector.dx;
            this.dy = this.dy - vector.dy;
    		this.sizeCacheReady = false;
        }

    	limit(number)
    	{
    		/*
    		if (this.sizeCacheReady)
    		{
    			if (this.sizeCache > number)
    			{
    				this.normalize();
    				this.multiply(number)
    				this.sizeCache = number;
    			}
    		}
    		else
    		{
    			let s = this.size2();
    			if (s > number * number)
    			{
    				this.normalize();
    				this.multiply(number);
    				this.sizeCache = number;
    				this.sizeCacheReady = true;
    			}
    		}
    		*/

    	}

    	normalize()
    	{
    		if (!this.sizeCacheReady)
    		{
    			this.sizeCache = Math.sqrt(this.dx * this.dx + this.dy * this.dy);
    			this.sizeCacheReady = true;
    		}

            if (this.sizeCache > 0)
            {
                this.divide(this.sizeCache);
    			this.sizeCache = 1;
            }
    	}

        getNormal()
        {
            var n = new Vector(-this.dy, this.dx);
            n.normalize();
            return n;
        }

    	//static version - returns new vector
    	static subtract(v1, v2)
    	{
    		return new Vector(v1.dx - v2.dx, v1.dy - v2.dy)
    	}

        static distance(v1, v2)
        {
            return Vector.subtract(v2, v1).size();
        }

    	static distance2(v1, v2)
    	{
    		return Vector.subtract(v2, v1).size2();
    	}

        static dot(v1, v2)
        {
            return v1.dx * v2.dx + v1.dy * v2.dy;
        }

        static reflect(v1, n)
        {
            var r = new Vector(n.dx, n.dy);
            r.multiply(2 * Vector.dot(v1, n));
            return Vector.subtract(v1, r);
        }

    }
