
function applyGravity(gameObjects, deltaTime)
{
    //for each pair of objects, the gravitational force only needs to be calculated
    //once, as the force is equal and opposite for the respective objects.

    let F;
    let a;
    let hyp;
    let hyp2;
    let mult;
    let deadZone = 75; //10
    let G = 3000; //1000

    for (let i=0; i<gameObjects.length-1; i++)
    {
        if (!gameObjects[i].applyGravity || gameObjects[i].mass == 0)
        {
            continue;
        }

        for (let j=(i+1); j<gameObjects.length; j++)
        {

            if (!gameObjects[j].applyGravity || gameObjects[j].mass == 0)
            {
                continue;
            }

            vec = Vector.subtract(gameObjects[j].position, gameObjects[i].position);
            hyp2 = vec.size2();

            if (hyp2 < deadZone * deadZone)
            {
                //no gravity to be applied within this radius
                continue;
            }

            hyp = Math.sqrt(hyp2);

            //force F = G * (m1 * m2) / (r ^ 2)

            F = G * (gameObjects[j].mass * gameObjects[i].mass) / hyp2;

            //calculate force as a vector and apply to both objects,
            //inverted for the second.

            //multipler to apply to the vector in case of anti-gravity/gravity combinations
            mult = gameObjects[i].antiGravity ? -1 : 1 * gameObjects[j].antiGravity ? -1 : 1;

            vec.normalize();
            vec.multiply(F);
            vec.multiply(mult);
            gameObjects[i].applyForce(vec, deltaTime);
            vec.multiply(-1);
            gameObjects[j].applyForce(vec, deltaTime);

        }
    }

}

class Stars
{
    constructor(width, height, number, levelfactor, starStyle = 1, starSize = 1)
    {
        this.width = width;
        this.height = height;
        this.number = number;
        this.levelfactor = levelfactor;

        this.starStyle = starStyle; //1 white, 2 twinky red, 3 twin peaks
        this.starSize = starSize;

        this.starsX = [];
        this.starsY = [];

        for (let i = 0; i < number; i++) {
            this.starsX.push(Math.random() * width / this.levelfactor);
            this.starsY.push(Math.random() * height / this.levelfactor);
        }
    }

    draw (shipX, shipY, canvasContext)
    {

        var canvasX;
        var canvasY;
        for (let i = 0; i < this.number; i++) {
            canvasX = this.mod((this.starsX[i] - (shipX + this.width / 2)) * this.levelfactor, this.width);
            canvasY = this.mod((this.starsY[i] - (shipY + this.height / 2)) * this.levelfactor, this.height);

            //normal star mode

            switch (this.starStyle)
            {
                case 1: //white
                    canvasContext.strokeStyle = canvasRGBA(255,255,255,this.levelfactor * 10); //0.05
                    break;
                case 2: //red sparkly
                    canvasContext.strokeStyle = canvasRGBA(255,Math.random()*255,Math.random()*255,this.levelfactor * 1); //1
                    break;
                case 3: //twin peaks - done in next section
                    break;
            }
            canvasContext.lineWidth = this.levelfactor * this.starSize;

            if (this.starStyle == 1 || this.starStyle == 2)
            {
                canvasContext.strokeRect(canvasX, canvasY, 1, 1);
            }
            else if (this.starStyle == 3)
            {
                //twin peaks
                canvasContext.strokeStyle = canvasRGBA(255,0,0,this.levelfactor * 0.05); //1
                canvasContext.lineWidth = this.levelfactor * this.starSize; //10
                canvasContext.beginPath();
                canvasContext.moveTo(canvasX, 0);
                canvasContext.lineTo(canvasX, this.height);
                canvasContext.stroke();

                canvasContext.strokeStyle = canvasRGBA(128,0,0,this.levelfactor * 0.03); //1
                canvasContext.lineWidth = this.levelfactor * this.starSize / 10; //1
                canvasContext.beginPath();
                canvasContext.moveTo(0, canvasY);
                canvasContext.lineTo(this.width, canvasY);
                canvasContext.stroke();
            }

        }

    }

    mod(n, m) {
        return ((n % m) + m) % m;
    }
}

class Reverb
{
    constructor(duration, decay, audioContext)
    {
        this.audioContext = audioContext;

        this.gainInput = audioContext.createGain();

        this.convolver = audioContext.createConvolver();
        this.gainConvolverDry = audioContext.createGain();
        this.gainConvolverWet = audioContext.createGain();
        this.gainFinal = audioContext.createGain();

        this.convolver.buffer = impulseResponse(duration, decay, false);

        this.gainInput.connect(this.convolver);
        this.gainInput.connect(this.gainConvolverDry);
        this.convolver.connect(this.gainConvolverWet);
        this.gainConvolverWet.connect(this.gainFinal);
        this.gainConvolverDry.connect(this.gainFinal);
    }

    setDryWet(dry, wet)
    {
        this.gainConvolverDry.gain.linearRampToValueAtTime(dry, this.audioContext.currentTime + 0.001);
        this.gainConvolverWet.gain.linearRampToValueAtTime(wet, this.audioContext.currentTime + 0.001);
    }

    connect(audioNode)
    {
        this.gainFinal.connect(audioNode);
    }

    getInput()
    {
        return this.gainInput;
    }
}

class LeftToRightTrack
{
    constructor(width, segmentLength, baseHeight, startX, startY, trackNoisyPer = 10)
    {
        this.runningCentre = [];
        this.topTrack = [];
        this.bottomTrack = [];
        this.topNoisy = [];
        this.bottomNoisy = [];
        this.topHitCount = [];
        this.bottomHitCount = [];

        var currentPos = new Vector(startX, startY);
        var currentDir = 0; //pointing right
        var height;

        this.topTrack.push(new Vector(startX, startY));
        this.bottomTrack.push(new Vector(startX, startY));

        while (currentPos.dx - startX < width)
        {
            height = baseHeight + Math.random() * (baseHeight / 8);
            this.runningCentre.push(new Vector(currentPos.dx, currentPos.dy));
            this.topTrack.push(new Vector(
                currentPos.dx + Math.cos(currentDir - Math.PI / 2) * height,
                currentPos.dy + Math.sin(currentDir - Math.PI / 2) * height
            ));
            this.bottomTrack.push(new Vector(
                currentPos.dx + Math.cos(currentDir + Math.PI / 2) * height,
                currentPos.dy + Math.sin(currentDir + Math.PI / 2) * height
            ));

            currentDir = currentDir + Math.random() * (2 * Math.PI / 32) - (2 * Math.PI / 64);
            if (currentDir < -(2 * Math.PI / 5)) currentDir = -(2 * Math.PI / 5);
            if (currentDir > (2 * Math.PI / 5)) currentDir = (2 * Math.PI / 5);

            currentPos.dx = currentPos.dx + Math.cos(currentDir) * segmentLength;
            currentPos.dy = currentPos.dy + Math.sin(currentDir) * segmentLength;

            this.topHitCount.push(0);
            this.bottomHitCount.push(0);

            if (trackNoisyPer == 0)
            {
                this.topNoisy.push(false);
                this.bottomNoisy.push(false);
            }
            else {
                this.topNoisy.push(Math.random() * trackNoisyPer < 1);
                this.bottomNoisy.push(Math.random() * trackNoisyPer < 1);
            }
        }

        var lastTop = this.topTrack[this.topTrack.length - 1];
        var lastBottom = this.bottomTrack[this.bottomTrack.length - 1];
        this.topTrack.push(new Vector(lastBottom.dx, lastBottom.dy));
        this.bottomTrack.push(new Vector(lastBottom.dx, lastBottom.dy));


    }

    draw(canvasContext)
    {
        var rv = 4;

        canvasContext.strokeStyle = canvasRGBA(0,255,255,255);
        canvasContext.fillStyle = canvasRGBA(0, 0, 0, 0.6);
        canvasContext.lineWidth = 4; //2
        //canvasContext.fillStyle = canvasRGBA(255,0,0,255);

        //top
        canvasContext.beginPath();
        canvasContext.moveTo(this.topTrack[0].dx + Math.random() * rv - rv/2, this.topTrack[0].dy + Math.random() * rv - rv/2);
        for (var i = 1; i < this.topTrack.length; i++) {
            canvasContext.lineTo(this.topTrack[i].dx + Math.random() * rv - rv/2, this.topTrack[i].dy + Math.random() * rv - rv/2);
        }
        canvasContext.lineTo(this.topTrack[this.topTrack.length-2].dx, this.topTrack[this.topTrack.length-2].dy);
        canvasContext.lineTo(this.topTrack[this.topTrack.length-2].dx, -100000);
        canvasContext.lineTo(this.topTrack[0].dx, -100000);
        canvasContext.lineTo(this.topTrack[0].dx, this.topTrack[0].dy);

        canvasContext.closePath();
        canvasContext.stroke();
        canvasContext.fill();

        //bottom
        canvasContext.beginPath();
        canvasContext.moveTo(this.bottomTrack[0].dx + Math.random() * rv - rv/2, this.bottomTrack[0].dy + Math.random() * rv - rv/2);
        for (var i = 1; i < this.bottomTrack.length; i++) {
            canvasContext.lineTo(this.bottomTrack[i].dx + Math.random() * rv - rv/2, this.bottomTrack[i].dy + Math.random() * rv - rv/2);
        }
        canvasContext.lineTo(this.bottomTrack[this.bottomTrack.length-1].dx, 100000);
        canvasContext.lineTo(this.bottomTrack[0].dx, 100000);
        canvasContext.lineTo(this.bottomTrack[0].dx, this.bottomTrack[0].dy);

        canvasContext.closePath();
        canvasContext.stroke();
        canvasContext.fill();

        //noisy segments & start/finish
        rv = 10;
        for (var i = 0; i < this.topTrack.length - 1; i++) {
            if (i == 0 || i == this.topTrack.length - 2)
            {
                canvasContext.strokeStyle = canvasRGBA(0,255,255,0.95);
                canvasContext.lineWidth = 16;
            }
            else {
                canvasContext.strokeStyle = canvasRGBA(255,128,128,0.9);
                canvasContext.lineWidth = 8;
            }
            if ((i == 0 || i == this.topTrack.length - 2) || this.topNoisy[i])
            {
                canvasContext.beginPath();
                canvasContext.moveTo(this.topTrack[i].dx + Math.random() * rv - rv/2, this.topTrack[i].dy + Math.random() * rv - rv/2);
                canvasContext.lineTo(this.topTrack[i+1].dx + Math.random() * rv - rv/2, this.topTrack[i+1].dy + Math.random() * rv - rv/2);
                canvasContext.stroke();
            }
            if ((i == 0) || this.bottomNoisy[i])
            {
                canvasContext.beginPath();
                canvasContext.moveTo(this.bottomTrack[i].dx + Math.random() * rv - rv/2, this.bottomTrack[i].dy + Math.random() * rv - rv/2);
                canvasContext.lineTo(this.bottomTrack[i+1].dx + Math.random() * rv - rv/2, this.bottomTrack[i+1].dy + Math.random() * rv - rv/2);
                canvasContext.stroke();
            }
        }

    }
}



class GameObject
{
    constructor(params)
    {
        this.direction = params.direction;
        this.position = new Vector(params.posX, params.posY);
        this.size = new Vector(params.sizeX, params.sizeY);
        this.speed = new Vector(params.speedX, params.speedY);
        this.deltaSpeed = new Vector(0,0);
        this.visible = true;
        this.dead = false;
        this.applyGravity = false;
        this.mass = 0;
        this.antiGravity = false;

        this.previousPosition = new Vector(params.posX, params.posY);
    }

    rotate(deltaRadians)
    {
        this.direction += deltaRadians;
        this.direction = this.direction % (2 * Math.PI);
    }

    update(deltaTime)
    {
        this.previousPosition.dx = this.position.dx;
        this.previousPosition.dy = this.position.dy;

        this.deltaSpeed.set(this.speed.dx, this.speed.dy);
        this.deltaSpeed.multiply(deltaTime);
        this.position.add(this.deltaSpeed);
    }

    isCrossingLine(startX, startY, endX, endY)
    {
        return intersects(
            this.previousPosition.dx, this.previousPosition.dy,
            this.position.dx, this.position.dy,
            startX, startY,
            endX, endY
        );
    }

    wrap(minX, minY, maxX, maxY, kill)
    {
        if (this.position.dx < minX)
        {
            if (kill) this.dead = true;
            else this.position.dx += (maxX - minX);
        }
        else if (this.position.dx > maxX)
        {
            if (kill) this.dead = true;
            else this.position.dx -= (maxX - minX);
        }

        if (this.position.dy < minY)
        {
            if (kill) this.dead = true;
            else this.position.dy += (maxY - minY);
        }
        else if (this.position.dy > maxY)
        {
            if (kill) this.dead = true;
            else this.position.dy -= (maxY - minY);
        }
    }

    draw(canvasContext)
    {

    }

    applyForce(F, deltaTime)
    {
        // apply acceleration
        // F = ma
        // a = F/m (apply to each component of the force vector)

        var finalF = new Vector(F.dx, F.dy);
        if (this.mass == 0) return;
        finalF.divide(this.mass);
        finalF.multiply(deltaTime);

        this.speed.add(finalF);
    }

}

class MonoStrip extends GameObject
{
    constructor(params)
    {
        super(params);

        this.makingMonos = false;
        this.shouldMakeMonoThisFrame = false;

        this.monoPeriod = 0.25;
        this.timeToNextMono = this.monoPeriod;

    }

    update(deltaTime)
    {
        super.update(deltaTime);

        this.shouldMakeMonoThisFrame = false;

        if (this.makingMonos)
        {
            this.timeToNextMono += deltaTime;
            if (this.timeToNextMono >= this.monoPeriod)
            {
                this.shouldMakeMonoThisFrame = true; //dealt with in the main game loop
                this.timeToNextMono = 0.0;
            }
        }
        else {
            this.timeToNextMono = this.monoPeriod; //this is so it creates one immediately
        }

    }

    draw(canvasContext)
    {
        var a = this.makingMonos ? 0.6 : 0.2;
        canvasContext.fillStyle = canvasRGBA(255,255,0, a + Math.random() * 0.02 - 0.01);
        canvasContext.strokeStyle = canvasRGBA(255,255,0, a + 0.2 + Math.random() * 0.2 - 0.1);
        canvasContext.lineWidth = 1;

        canvasContext.fillRect
        (
            this.position.dx - this.size.dx / 2 + Math.random() * 2 - 1,
            this.position.dy - this.size.dy / 2 + Math.random() * 2 - 1,
            this.size.dx + Math.random() * 4 - 2,
            this.size.dy + Math.random() * 4 - 2
        );

        canvasContext.strokeRect
        (
            this.position.dx - this.size.dx / 2 + Math.random() * 4 - 2,
            this.position.dy - this.size.dy / 2 + Math.random() * 4 - 2,
            this.size.dx + Math.random() * 4 - 2,
            this.size.dy + Math.random() * 4 - 2
        );
    }

}

class TextObject extends GameObject
{
    constructor(params, text)
    {
        super(params);
        this.text = text;
    }

    update(deltaTime)
    {
        super.update(deltaTime);
    }

    draw(canvasContext)
    {
        canvasCtx.fillStyle = canvasRGBA(255, 255, 255, 0.9);
        canvasCtx.font = Math.floor(this.size.dy / 2) + 'px sans-serif';
        canvasCtx.fillText(this.text, this.position.dx, this.position.dy + this.size.dy / 2, this.size.dx);
    }

}

class AudioGameObject extends GameObject
{
    constructor(params, audioContext)
    {
        super(params);
        this.audioContext = audioContext;

        this.gainInput = audioContext.createGain();
        this.gainFinal = audioContext.createGain(); //what subclasses should connect to.
                            //even though really it's not final as followed by panners

        this.panner = audioContext.createStereoPanner();
        this.pannerGain = audioContext.createGain();

        this.panner.pan.value = 0.0;
        this.pannerGain.gain.value = 1.0;

        this.gainFinal.connect(this.panner);
        this.panner.connect(this.pannerGain);
    }

    connect(audioNode)
    {
        this.pannerGain.connect(audioNode);
    }

    disconnect(audioNode)
    {
        try {
            this.pannerGain.disconnect(audioNode);
        } catch (e) {
            //no connection existed - carry on
        }
    }

    getInput()
    {
        return this.gainInput;
    }

    update(deltaTime)
    {
        super.update(deltaTime);
    }

    draw(canvasContext)
    {

    }

    updatePan(relativeTo, maxDist)
    {
        let maxDist2 = maxDist * maxDist;
        let sub = Vector.subtract(this.position, relativeTo.position);
        let dist2 = sub.size2();
        let newGain;

        if (dist2 > maxDist2)
        {
            this.pannerGain.gain.linearRampToValueAtTime(0.0, this.audioContext.currentTime + 0.001);
        }
        else {
            newGain = Math.pow(1 - dist2 / maxDist2, 5);
            this.pannerGain.gain.linearRampToValueAtTime(newGain, this.audioContext.currentTime + 0.001);
            let newPan = sub.dx / (maxDist / 4); //100
            if (newPan < -1) newPan = -1;
            if (newPan > 1) newPan = 1;
            this.panner.pan.linearRampToValueAtTime(newPan, this.audioContext.currentTime + 0.001);
        }
    }

}

class MonoThing extends AudioGameObject
{
    constructor(params, audioContext)
    {
        super(params, audioContext);

        this.useAnalyser = true;

        this.killAfterNextNote = false;

        this.playNoteActive = true;
        this.playNoteTimeToNext = 0.0;
        this.playNotePeriod = 3.0;
        this.note = 49;

        this.osc = audioContext.createOscillator();
        this.env = new Envelope(0.9, 0.5, 0.0, 0.7); //0.01, 0.5, 0.0, 0.2);
        this.envGain = audioContext.createGain();

        this.osc.type = (Math.random() * 2 < 1) ? 'triangle' : 'sine';
        this.envGain.gain.value = 0.0;

        this.osc.connect(this.envGain);
        this.envGain.connect(this.gainFinal);
        this.osc.start(0);

        if (this.useAnalyser)
        {
            this.lastUpdateVolumeAverage = 0;

            this.analyser = audioContext.createAnalyser();
            this.analyser.fftSize = 128;
            this.analyserBuffer = new Uint8Array(this.analyser.fftSize);

            this.gainFinal.connect(this.analyser);
        }


    }

    playNote(note)
    {
        this.osc.frequency.linearRampToValueAtTime(freq12TET(note, note, 0), audioContext.currentTime + 0.001);
        this.env.applyOn(this.envGain.gain, this.audioContext);
    }

    stopNote()
    {
        this.env.applyOff(this.envGain.gain, this.audioContext);
    }

    update(deltaTime)
    {
        if (!this.killAfterNextNote)
            super.update(deltaTime);

        if (this.dead) return;

        if (this.playNoteActive)
        {
            this.playNoteTimeToNext += deltaTime;
            if (this.playNoteTimeToNext >= this.playNotePeriod)
            {
                this.playNote(this.note);
                this.playNoteTimeToNext = 0.0;
                if (this.killAfterNextNote) this.dead = true;
            }
        }

        if (this.useAnalyser)
        {
            this.analyser.getByteTimeDomainData(this.analyserBuffer);

            //average the buffer to get a draw size;
            var s = 0;
            for (let i=0; i < this.analyserBuffer.length; i++)
            {
                s += Math.abs(this.analyserBuffer[i] - 128);
            }
            s = s / this.analyserBuffer.length;
            s = s / 128.0 * 3;

            this.lastUpdateVolumeAverage = s;
        }

        this.rotate(Math.PI * 2 * deltaTime * 0.5 + s);

    }

    draw(canvasContext)
    {
        if (this.dead) return;

        var s;
        if (this.useAnalyser)
        {
            s = 10 + this.lastUpdateVolumeAverage * 750; //400 //1000
        }
        else {
            s = this.size.dx;
        }

        canvasContext.strokeStyle = canvasRGBA(s * 10 % 255,s * 10 % 255,255,255);
        canvasContext.lineWidth = 4; //5
        //canvasContext.fillStyle = canvasRGBA(255,0,0,255);

        canvasContext.strokeRect
        (
            this.position.dx - s / 2,
            this.position.dy - s / 2,
            s,
            s
        );
        /*
        canvasContext.fillRect
        (
            this.position.dx - this.size.dx * s / 2,
            this.position.dy - this.size.dy * s / 2,
            this.size.dx * s,
            this.size.dy * s
        );
        */

    }

}

/* template for individual audio game objects
class ... extends AudioGameObject
{
    constructor(params, audioContext)
    {
        super(params, audioContext);
    }

    update(deltaTime)
    {
        super.update(deltaTime);

        if (this.dead) return;
    }

    draw(canvasContext)
    {
        if (this.dead) return;
    }

}
*/

class GravityShip extends AudioGameObject
{
    constructor(params, audioContext)
    {
        super(params, audioContext);
    }

    update(deltaTime)
    {
        super.update(deltaTime);

        if (this.dead) return;
    }

    draw(canvasContext)
    {
        if (this.dead) return;

        canvasContext.strokeStyle = canvasRGBA(255,255,255,0.2 + this.mass / 250);
        canvasContext.lineWidth = 2;
        //canvasContext.fillStyle = canvasRGBA(255,0,0,255);

        canvasCircle(canvasContext, this.position.dx, this.position.dy, this.mass, false)

        var r = 0;
        var r2 = 0;
        if (this.mass == 0)
        {
            canvasContext.strokeStyle = canvasRGBA(255,255,255,1);
        }
        else {
            canvasContext.strokeStyle = canvasRGBA(255,255,255,0.22);
            r = Math.random() * 5 - 2.5;
            r2 = Math.random() * 5 - 2.5;
        }

        canvasContext.beginPath();

        canvasContext.moveTo(this.position.dx - this.size.dx / 3 + r, this.position.dy - this.size.dy / 3 + r2);
        canvasContext.lineTo(this.position.dx + this.size.dx / 2 + r, this.position.dy + r2);
        canvasContext.lineTo(this.position.dx - this.size.dx / 3 + r, this.position.dy + this.size.dy / 3 + r2);
        canvasContext.closePath();

        canvasContext.stroke();


    }

}

class EchoBox extends AudioGameObject
{
    constructor(params, audioContext)
    {
        super(params, audioContext);

        this.lastUpdateVolumeAverage = 0;
        this.silentThreshold = 0.05;
        this.silentTimeToDeath = 30;
        this.silentTimeToDying = 2;
        this.silentTime = 0;

        this.warningTimeToNext = 0;
        this.warningPeriod = 0.33;

        this.delay = audioContext.createDelay(2.0);
        this.gainFeedback = audioContext.createGain();
        this.filterFeedback = audioContext.createBiquadFilter();

        this.delay.delayTime.value = 0.75; //0.75 //2.0
        this.gainFeedback.gain.value = 0.94; //0.88 //0.98
        // this.filterFeedback.frequency.value = 1000; //22k
        // this.filterFeedback.type = "lowpass";
        // this.filterFeedback.Q.value = 0;

        this.warningOsc = audioContext.createOscillator();
        this.warningEnv = new Envelope(0.02, 0.05, 0.0, 0.2, 0.02); //0.04, 0.1, 0.0, 0.2, 0.02
        this.warningGain = audioContext.createGain();
        this.warningOsc.type = 'triangle';
        this.warningOsc.frequency.value = 1800;
        this.warningGain.gain.value = 0.0;

        this.gainInput.connect(this.delay);
        this.delay.connect(this.gainFinal);
        this.delay.connect(this.gainFeedback);
        this.gainFeedback.connect(this.delay); //this.filterFeedback;
        //this.filterFeedback.connect(this.delay);

        this.warningOsc.connect(this.warningGain);
        this.warningGain.connect(this.panner);

        this.analyser = audioContext.createAnalyser();
        this.analyser.fftSize = 128;
        this.analyserBuffer = new Uint8Array(this.analyser.fftSize);

        this.gainFinal.connect(this.analyser);

        this.warningOsc.start(0);
    }

    update(deltaTime)
    {
        super.update(deltaTime);

        if (this.dead) return;

        this.rotate(Math.PI * 2 * deltaTime * 0.1);
        //this.direction += Math.PI * 2 * deltaTime * 50;
        //this.direction = this.direction % (2 * Math.PI);

        this.analyser.getByteTimeDomainData(this.analyserBuffer);

        //average the buffer to get a draw size;
        var s = 0;
        for (let i=0; i < this.analyserBuffer.length; i++)
        {
            s += Math.abs(this.analyserBuffer[i] - 128);
        }
        s = s / this.analyserBuffer.length;
        s = s / 128.0 * 3;

        this.lastUpdateVolumeAverage = s;

        if (s < this.silentThreshold)
        {
            this.silentTime += deltaTime;
            if (this.silentTime > this.silentTimeToDeath)
            {
                this.dead = true;
                this.gainFinal.gain.value = 0.0;
                this.warningOsc.frequency.linearRampToValueAtTime(112.5, this.audioContext.currentTime + 0.01);
                this.warningEnv.applyOn(this.warningGain.gain, this.audioContext);
            }
        }
        else {
            this.silentTime = 0.0;
        }

        if (this.silentTime > this.silentTimeToDying)
        {
            this.warningTimeToNext += deltaTime;
            if (this.warningTimeToNext >= this.warningPeriod)
            {
                this.warningEnv.applyOn(this.warningGain.gain, this.audioContext);
                this.warningTimeToNext = 0.0;
            }
        }
        else {
            this.warningTimeToNext = 0.0;
        }
    }

    draw(canvasContext)
    {
        if (this.dead) return;

        let s = this.lastUpdateVolumeAverage;
        if (s > 1) s = 1;

        if (this.silentTime > this.silentTimeToDying)
        {
            canvasContext.strokeStyle = canvasRGBA(255,255,0,255);
            canvasContext.lineWidth = 1;
            canvasContext.fillStyle = canvasRGBA(255,0,0,255);
        }
        else
        {
            canvasContext.strokeStyle = canvasRGBA(0,255,255,255);
            canvasContext.lineWidth = 1;
            canvasContext.fillStyle = canvasRGBA(0,255,255,128);
        }

        canvasContext.strokeRect
        (
            this.position.dx - this.size.dx / 2,
            this.position.dy - this.size.dy / 2,
            this.size.dx,
            this.size.dy
        );

        canvasContext.fillRect
        (
            this.position.dx - this.size.dx * s / 2,
            this.position.dy - this.size.dy * s / 2,
            this.size.dx * s,
            this.size.dy * s
        );
    }
}


//this class has no value at the moment (the params object could be used directly)
//but might be useful at some point to keep track of progress, completion etc.
class GameLevel
{
    constructor(params)
    {
        this.starStyle = params.starStyle; //1 white, 2 twinky red, 3 twin peaks
        this.starSize = params.starSize; //1 small, 2 large
        this.useTrack = params.useTrack;
        this.trackWidth = params.trackWidth;
        this.trackHeight = params.trackHeight;
        this.nonTrackTime = params.nonTrackTime; //time you need to stay alive for in non-track level
        this.initialMonos = params.initialMonos;
        this.monoWrap = params.monoWrap;
        this.monoClearToComplete = params.monoClearToComplete; //bool = clear all monos to complete level
        this.trackNoisyPer = params.trackNoisyPer;
        this.monoStripPer = params.monoStripPer;
        this.music = params.music; //TODO
        this.shipTopSpeed = params.shipTopSpeed;
        this.echoInitialFeedback = params.echoInitialFeedback;
        this.text = params.text;
    }
}

function getLevels()
{
    var levels = [];

    //level 0 - startup/interstitial
    var params = {
        starStyle: 3,
        starSize: 2,
        useTrack: false,
        trackWidth: 0,
        trackHeight: 0,
        nonTrackTime: 15,
        initialMonos: 0,
        monoWrap: true,
        monoClearToComplete: false,
        trackNoisyPer: 0,
        monoStripPer: 0,
        music: 2,
        shipTopSpeed: 800,
        echoInitialFeedback: 0.94,
        text: "Click to start"
    }
    levels.push(new GameLevel(params));

    //level 1
    var params = {
        starStyle: 1,
        starSize: 1,
        useTrack: false,
        trackWidth: 0,
        trackHeight: 0,
        nonTrackTime: 0,
        initialMonos: 10,
        monoWrap: true,
        monoClearToComplete: true,
        trackNoisyPer: 0,
        monoStripPer: 0,
        music: 2,
        shipTopSpeed: 400,
        echoInitialFeedback: 0.94,
        text: "ECHO will follow you - it needs sound to live, so make it collect all the sounds..."
    }
    levels.push(new GameLevel(params));

    var params = {
        starStyle: 2,
        starSize: 8,
        useTrack: true,
        trackWidth: 54000,
        trackHeight: 300,
        nonTrackTime: 0,
        initialMonos: 0,
        monoWrap: false,
        monoClearToComplete: false,
        trackNoisyPer: 10,
        monoStripPer: 0,
        music: 3,
        shipTopSpeed: 300,
        echoInitialFeedback: 0.95, //0.95
        text: "Get ECHO to end of the track! Some of the walls might be noisy..."
    }
    levels.push(new GameLevel(params));

    var params = {
        starStyle: 3,
        starSize: 10,
        useTrack: false,
        trackWidth: 0,
        trackHeight: 0,
        nonTrackTime: 120,
        initialMonos: 0,
        monoWrap: false,
        monoClearToComplete: false,
        trackNoisyPer: 0,
        monoStripPer: 0,
        music: 3,
        shipTopSpeed: 500,
        echoInitialFeedback: 0.95,
        text: "Use UP to resonate ship near ECHO to feed it sound & survive 2 mins!"
    }
    levels.push(new GameLevel(params));

    var params = {
        starStyle: 2,
        starSize: 2,
        useTrack: false,
        trackWidth: 0,
        trackHeight: 0,
        nonTrackTime: 0,
        initialMonos: 10,
        monoWrap: true,
        monoClearToComplete: true,
        trackNoisyPer: 0,
        monoStripPer: 50,
        music: 3,
        shipTopSpeed: 500,
        echoInitialFeedback: 0.8,
        text: "Make ECHO collect all sounds whilst staying alive. Yellow squares may help / hinder..."
    }
    levels.push(new GameLevel(params));

    var params = {
        starStyle: 3,
        starSize: 20,
        useTrack: true,
        trackWidth: 34000,
        trackHeight: 300,
        nonTrackTime: 0,
        initialMonos: 0,
        monoWrap: true,
        monoClearToComplete: false,
        trackNoisyPer: 0,
        monoStripPer: 10, //40
        music: 1,
        shipTopSpeed: 500, //300
        echoInitialFeedback: 0.94,
        text: "Get ECHO to end of the track!"
    }
    levels.push(new GameLevel(params));





    return levels;
}
